"""
Contains the main interface, used to parse arguments and run the program.
"""

import argparse
import os
import re
from typing import Dict, List, Set

from argsearch import commands, ranges, strategies


def positive_int(arg: str) -> int:
    value = int(arg)
    if value < 1:
        raise argparse.ArgumentTypeError("Value must be a positive integer.")
    return value


def parse_range_args(
    range_args: List[str], templates: Set[str]
) -> Dict[str, ranges.Range]:
    """
    Parse the provided range arguments into Range objects.

    Parameters
    ----------
    range_args
        A list of range arguments provided on the command line.
    templates
        A list of template names to be filled in the command string.

    Returns
    -------
    Dict[str, Range]
        A mapping from template names to their specified ranges.
    """
    usage = "..."
    for template in templates:
        usage += f" --{template} (low high | A B C [D ...])"

    range_parser = argparse.ArgumentParser(
        "ranges",
        description="One range specifier per template in the input command. "
        "If two numbers (int or float) are provided, they are treated as the min and "
        "max of a numeric range. Otherwise, they are treated as separate categories.",
        add_help=True,
        usage=usage,
        allow_abbrev=False,
    )
    range_group = range_parser.add_argument_group("templates specified")

    try:
        for template in templates:
            range_group.add_argument(
                f"--{template}", action=ranges.TemplateRange, nargs="*", required=True,
            )
        parsed_ranges = range_parser.parse_args(range_args)
    except SystemExit:
        os._exit(2)

    return vars(parsed_ranges)


def get_template_names(command_string: str) -> Set[str]:
    """
    Get the names of all of the bracketed templates in a command string.

    Parameters
    ----------
    command_string
        The string to read templates from.

    Returns
    -------
    List[str]
        A list of the templates in the command string, without brackets.
    """
    templates = re.findall("\{.+?\}", command_string)
    names = {template.strip("{}") for template in templates}
    return names


def main():
    base_parser = argparse.ArgumentParser(
        description="Run the same command multiple times with different values for its "
        "arguments.",
    )
    base_parser.add_argument(
        "--num-workers",
        type=positive_int,
        default=0,
        metavar="N",
        help="if provided, split work among N worker processes",
    )
    base_parser.add_argument(
        "--output-json",
        action="store_true",
        help="capture output and format it as JSON instead of streaming to terminal",
    )

    base_parser.add_argument(
        "--disable-bar", action="store_true", help="disable the progress bar"
    )

    strategy_parsers = base_parser.add_subparsers(
        title="strategy", description="the search strategy to use", dest="strategy",
    )
    # Workaround for older Python versions; see
    # https://stackoverflow.com/questions/23349349/argparse-with-required-subparser
    strategy_parsers.required = True

    random_parser = strategy_parsers.add_parser("random", help="random search")
    random_parser.add_argument(
        "trials", type=positive_int, help="number of random trials to run"
    )
    random_parser.set_defaults(strategy="random")

    grid_parser = strategy_parsers.add_parser("grid", help="grid search")
    grid_parser.add_argument(
        "divisions",
        type=positive_int,
        help="number of ways to divide each numeric interval",
    )
    grid_parser.set_defaults(strategy="grid")

    repeat_parser = strategy_parsers.add_parser("repeat", help="repeat a command")
    repeat_parser.add_argument(
        "repeats", type=positive_int, help="number of repeats to run"
    )
    repeat_parser.set_defaults(strategy="repeat")
    repeat_parser.add_argument("command", help="the command to run")

    for subparser in [random_parser, grid_parser]:
        subparser.add_argument(
            "command",
            help="the command to run, including at least one bracketed template",
        )
        subparser.add_argument(
            "ranges",
            nargs=argparse.REMAINDER,
            help="a numeric or categorical range for each template in the command",
        )

    base_args = base_parser.parse_args()
    templates = get_template_names(base_args.command)

    if base_args.strategy != "repeat":
        if not templates:
            raise ValueError(
                "At least one template must be provided with the "
                f"'{base_args.strategy}' strategy."
            )
        parsed_ranges = parse_range_args(base_args.ranges, templates)

    if base_args.strategy == "random":
        command_strings = strategies.random(
            base_args.command, parsed_ranges, base_args.trials
        )
    elif base_args.strategy == "grid":
        command_strings = strategies.grid(
            base_args.command, parsed_ranges, base_args.divisions
        )
    elif base_args.strategy == "repeat":
        command_strings = strategies.repeat(base_args.command, base_args.repeats)
    else:
        raise ValueError(f"Unrecognized strategy: {base_args.strategy}.")

    commands.run_commands(
        command_strings,
        base_args.output_json,
        base_args.num_workers,
        base_args.disable_bar,
    )



"""
Defines Ranges, which specify how to sample arguments of various types.
"""

import abc
import argparse
from typing import List
import random

import numpy as np


class Range(abc.ABC):
    """
    A base class for ranges that can be specified for templated arguments.
    """

    @abc.abstractmethod
    def random_sample(self) -> str:
        """
        Get a random sample from this range.

        Returns
        -------
        str
            A value to be substituted into the command in place of a bracketed template.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def grid(self, divisions: int) -> List[str]:
        """
        Get the list of all possible values when divided into `divisions` pieces.

        Parameters
        ----------
        divisions
            How many axis divisions we're using for grid search.

        Returns
        -------
        List[str]
            The list of all values that will be searched over on this range.
        """
        raise NotImplementedError


class IntRange(Range):
    """
    A range of integral values between a minimum and a maximum.
    """

    def __init__(self, min_value: int, max_value: int):
        self.min_value = min_value
        self.max_value = max_value

    def random_sample(self) -> str:
        return str(random.randint(self.min_value, self.max_value))

    def grid(self, divisions: int) -> List[str]:
        divisions = min(divisions, self.max_value - self.min_value + 1)
        space = np.linspace(self.min_value, self.max_value, num=divisions, dtype=int)
        return list(map(str, space))


class FloatRange(Range):
    """
    A range of floating-point values between a minimum and a maximum.
    """

    def __init__(self, min_value: float, max_value: float):
        self.min_value = min_value
        self.max_value = max_value

    def random_sample(self) -> str:
        return str(random.uniform(self.min_value, self.max_value))

    def grid(self, divisions: int) -> List[str]:
        space = np.linspace(self.min_value, self.max_value, num=divisions, dtype=float)
        return list(map(str, space))


class CategoricalRange(Range):
    """
    A range containing a fixed set of categories.

    NOTE: this is not divided during grid search; every category will be searched over.
    If this behavior is not desirable, consider a random search.
    """

    def __init__(self, categories: List[str]):
        self.categories = categories

    def random_sample(self) -> str:
        return random.choice(self.categories)

    def grid(self, divisions: int) -> List[str]:
        return self.categories


class TemplateRange(argparse.Action):
    """
    An argparse Action to parse a template range argument into a Range object.
    """

    def __call__(self, parser, namespace, values, option_string):
        if len(values) < 2:
            raise argparse.ArgumentTypeError(
                "A template range requires 2 or more values."
            )

        if len(values) == 2:
            try:
                value_1 = int(values[0])
                value_2 = int(values[1])
                min_value = min(value_1, value_2)
                max_value = max(value_1, value_2)
                setattr(namespace, self.dest, IntRange(min_value, max_value))
                return
            except ValueError:
                pass

            try:
                value_1 = float(values[0])
                value_2 = float(values[1])
                min_value = min(value_1, value_2)
                max_value = max(value_1, value_2)
                setattr(namespace, self.dest, FloatRange(min_value, max_value))
                return
            except ValueError:
                pass

        setattr(namespace, self.dest, CategoricalRange(values))



"""
Functions to run user commands.
"""

import json
import multiprocessing
import subprocess
import sys
from typing import Any, Dict, List, Optional, Tuple

from tqdm import tqdm


def format_header(step: int, command: str):
    return f"--- [{step}] {command}"


def stream_command(command: str, step: int, monitor: tqdm) -> None:
    """
    Run a command string, streaming output to the stdout.

    Parameters
    ----------
    command
        The command string to run. Tokenized with the shell defaults.
    step
        Which step of the search we're on.
    monitor
        A handle to the parent progress bar.
    """
    monitor.write(format_header(step, command))
    process = subprocess.Popen(
        command, shell=True, stdout=subprocess.PIPE, encoding="utf-8",
    )

    assert process.stdout
    for line in process.stdout:
        monitor.write(line, end="")

    process.wait()


def capture_command(command: str, step: int, monitor: Optional[tqdm]) -> Dict[str, Any]:
    """
    Run a command string, capturing and formatting any output.

    Parameters
    ----------
    command
        The command string to run. Tokenized with the shell defaults.
    step
        Which step of the search we're on.
    monitor
        An optional handle to the parent progress bar.

    Returns
    -------
    Dict[str, str]
        The results of evaluating the command with substitution.
    """
    if monitor:
        monitor.set_description(command)
    command_result = subprocess.run(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True,
        encoding="utf-8",
    )

    return {
        "step": step,
        "command": command,
        "stdout": command_result.stdout,
        "stderr": command_result.stderr,
        "returncode": command_result.returncode,
    }


def _capture_command_packed(args: Tuple[str, int, tqdm]) -> Dict[str, Any]:
    return capture_command(*args)


def run_commands(
    command_strings: List[str],
    output_json: bool = False,
    num_workers: int = 0,
    disable_bar: bool = False,
) -> None:
    """
    Run a list of command strings, streaming or formatting the output.

    Parameters
    ----------
    command_strings
        All command strings to run.
    output_json
        If True, collect output and print at the end, formatted as json.
        Otherwise (default), stream output to stdout as it arrives.
    num_workers
        If provided, use this many worker processes to run commands.
    disable_bar
        If True, disable the progress bar.
    """
    if num_workers > 0:
        process_pool = multiprocessing.Pool(
            num_workers, initializer=tqdm.set_lock, initargs=(tqdm.get_lock(),)
        )

        with tqdm(total=len(command_strings), disable=disable_bar) as monitor:
            args_packed = [(comm, i, None) for i, comm in enumerate(command_strings)]

            if output_json:
                outputs = []

            for output in process_pool.imap_unordered(
                _capture_command_packed, args_packed
            ):
                monitor.update()

                if output_json:
                    outputs.append(output)
                else:
                    header = format_header(output["step"], output["command"])
                    output_with_header = f'{header}\n{output["stdout"]}'
                    monitor.write(output_with_header, end="")
                    if output["stderr"]:
                        sys.stderr.write(output["stderr"])
                        sys.stderr.flush()

            if output_json:
                formatted = json.dumps(outputs)
                monitor.write(formatted)

        return

    with tqdm(command_strings, disable=disable_bar) as monitor:
        if output_json:
            outputs = [
                capture_command(command, step, monitor)
                for step, command in enumerate(monitor)
            ]
            formatted = json.dumps(outputs)
            monitor.write(formatted)
        else:
            for step, command in enumerate(monitor):
                stream_command(command, step, monitor)



"""
Defines strategies, which sample arguments from Ranges to get command strings to run.
"""


import itertools
from typing import Dict, List

from argsearch import ranges


def apply_substitutions(command: str, substitutions: Dict[str, str]) -> str:
    """
    Given a command string (with templates) and a substitution mapping, produce a
    concrete command string that can be run in a shell.

    Parameters
    ----------
    command
        A string to be executed as a subprocess, with "{arg}" bracketed templates.
    substitutions
        Maps from a bracketed template name to a value to substitute.

    Returns
    -------
    str
        A command string, containing `substitutions` plugged into templates in
        `command`, that can be run in a shell.
    """
    for template, value in substitutions.items():
        command = command.replace("{" + template + "}", value)
    return command


def random(
    templated_string: str, range_map: Dict[str, ranges.Range], trials: int
) -> List[str]:
    """
    Get a list of command strings by random sampling.

    Parameters
    ----------
    templated_string
        A string to be executed as a subprocess, with "{arg}" bracketed templates.
    range_map
        Maps from a template name to a range defining values for that template.
    trials
        How many random trials to run.

    Returns
    -------
    List[str]
        A list of command strings, of length `trials`.
    """

    def random_sample():
        substitutions = {name: rng.random_sample() for name, rng in range_map.items()}
        command_string = apply_substitutions(templated_string, substitutions)
        return command_string

    return [random_sample() for _ in range(trials)]


def grid(
    templated_string: str, range_map: Dict[str, ranges.Range], divisions: int
) -> List[str]:
    """
    Get a list of all command strings to run in a grid search.

    Parameters
    ----------
    templated_string
        A string to be executed as a subprocess, with "{arg}" bracketed templates.
    range_map
        Maps from a template name to a range defining values for that template.
    divisions
        How many slices to divide each numeric range into.

    Returns
    -------
    List[str]
        A list of command strings, for each point on the sampled grid.
    """
    template_names, template_ranges = zip(*range_map.items())
    grids = [rng.grid(divisions) for rng in template_ranges]
    combinations = list(itertools.product(*grids))

    command_strings = []
    for combination in combinations:
        substitutions = dict(zip(template_names, combination))
        command_string = apply_substitutions(templated_string, substitutions)
        command_strings.append(command_string)

    return command_strings


def repeat(command_string: str, repeats: int) -> List[str]:
    """
    Get a list of command strings by repetition.

    Parameters
    ----------
    command_string
        A string to be executed as a subprocess, without any templates.
    repeats
        How many repeats to run.

    Returns
    -------
    List[str]
        `command_string`, repeated `repeats` times.
    """
    return [command_string] * repeats



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['argsearch']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19.1,<2.0.0', 'tqdm>=4.49.0,<5.0.0']

entry_points = \
{'console_scripts': ['argsearch = argsearch:main']}

setup_kwargs = {
    'name': 'argsearch',
    'version': '0.2.1',
    'description': 'Run a command many times with different combinations of its inputs.',
    'long_description': '# argsearch\n`argsearch` is a simple and composable tool for running the same command many times with different combinations of arguments.\nIt aims to make random search and grid search easy for things like hyperparameter tuning and setting simulation parameters, while only requiring that your program accepts command line arguments in some form.\n\n## Example\n```\n$ argsearch grid 3 \'echo {a} {b}\' --a 0.0 1.5 --b X Y\n[\n{"args": {"a": "0.0", "b": "X"}, "command": "echo 0.0 X", "stdout": "0.0 X\\n", "stderr": "", "returncode": 0},\n{"args": {"a": "0.0", "b": "Y"}, "command": "echo 0.0 Y", "stdout": "0.0 Y\\n", "stderr": "", "returncode": 0},\n{"args": {"a": "0.75", "b": "X"}, "command": "echo 0.75 X", "stdout": "0.75 X\\n", "stderr": "", "returncode": 0},\n{"args": {"a": "0.75", "b": "Y"}, "command": "echo 0.75 Y", "stdout": "0.75 Y\\n", "stderr": "", "returncode": 0},\n{"args": {"a": "1.5", "b": "X"}, "command": "echo 1.5 X", "stdout": "1.5 X\\n", "stderr": "", "returncode": 0},\n{"args": {"a": "1.5", "b": "Y"}, "command": "echo 1.5 Y", "stdout": "1.5 Y\\n", "stderr": "", "returncode": 0}\n]\n```\n\n## Installation\n\n```\npip install argsearch\n```\n\n## Usage\n\n`argsearch` takes 3 kinds of arguments:\n - A **search strategy** (*random,* *grid,* or *repeat*) and its configuration:\n    - For *random*: **trials**, the number of random trials to run.\n    - For *grid*: **divisions**, the number of points to try in each numeric range.\n    - For *repeat*: **repeats**, the number of times to repeat the command.\n - A **command string** with **templates** designated by bracketed names (e.g. `\'python my_script.py --flag {value}\'`.\n -  A **range** for each template in the command string (e.g. `--value 1 100`).\n\nThen, `argsearch` runs the command string several times, each time replacing the templates with values from their associated ranges.\nNote that the *repeat* strategy does not admit templates or range arguments.\n\n### Search Strategies\n\nThree search strategies are currently implemented:\n - **Random search** samples uniformly randomly from specified ranges for a fixed number\n     of trials.\n - **Grid search** divides each numeric range into a fixed number of\n     evenly-spaced points and runs once for each possible combination of\n     inputs.\n - **Repeat** runs the same command a fixed number of times.\n\n### Ranges\nThree types of ranges are available:\n - **Floating-point ranges** are specified by a minimum and maximum floating-point value (e.g. `--value 0.0 1.0`).\n - **Integer ranges** are specified by a minimum and maximum integer (e.g. `--value 1 100`). Integer ranges are guaranteed to only yield integer values.\n - **Categorical ranges** are specified by a list of non-numeric categories, or more than two numbers (e.g. `--value A B C`, `--value 2 4 8 16`). Categorical ranges only draw values from the listed categories, and are not divided up during a grid search.\n\n### Output\n\nThe output is JSON, which can be wrangled with [jq](https://github.com/stedolan/jq) or other tools, or dumped to a file. The run is a list of mappings, one per command call, each of which has the following keys:\n - `args`: a mapping of argument names to values.\n - `command`: the command string with substitutions applied.\n - `stdout`: a string containing the command\'s stdout.\n - `stderr`: a string containing the command\'s stderr.\n - `returncode`: an integer return code for the command.\n',
    'author': 'Aidan Swope',
    'author_email': 'aidanswope@gmail.com',
    'maintainer': 'Aidan Swope',
    'maintainer_email': 'aidanswope@gmail.com',
    'url': 'https://github.com/maxwells-daemons/argsearch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
